import time
import pymongo
from pymongo import UpdateOne, DeleteOne
from pymongo.errors import ConnectionFailure

from . import utils
from .records import Record, DocumentCollection, ZwdbError

class ZWMongo(object):
    """Class defining a Mongo driver"""
    def __init__(self, db_url, **kwargs):
        # http://dochub.mongodb.org/core/connections
        o = utils.db_url_parser(db_url)
        p = o['props']
        self.dbcfg = {
            'host'      : o['host'],
            'port'      : o['port'] or 27017,
            'username'  : o['usr'],
            'password'  : o['pwd']
        }
        self.dbname = o['db']
        self.dbcfg.update(kwargs)
        cfg = {
            'maxPoolSize' : 100
        }
        for p in cfg:
            self.dbcfg[p] = self.dbcfg.get(p, cfg[p])

        self.client = None
        client = pymongo.MongoClient(**self.dbcfg)
        try:
            # The ismaster command is cheap and does not require auth.
            client.admin.command('ismaster')
            self.client = client
        except ConnectionFailure:
            raise ZwdbError('Server not available, {}:{}'.format(self.dbcfg['host'], self.dbcfg['port']))

    @property
    def pool_size(self):
        return self.client.max_pool_size

    @property
    def version(self):
        return pymongo.version
    
    @property
    def server_info(self):
        return self.client.server_info()

    def get_connection(self):
        db = self.client[self.dbname]
        return ZWMongoConnection(db)

    def close(self):
        if self.client:
            self.client.close()

    def lists(self):
        return self.client[self.dbname].list_collection_names()
    
    def find(self, coll, conds=None, projection=None, sort=None, limit=0, fetchall=False, **params):
        conn = self.get_connection()
        docs = conn.find(coll, conds, projection, sort, limit, fetchall, **params)
        if fetchall:
            conn.close()
        return docs

    def insert(self, coll, recs, ordered=False):
        with self.get_connection() as conn:
            rtn = conn.insert(coll, recs, ordered)
        return rtn

    def update(self, coll, recs, keyflds=None):
        with self.get_connection() as conn:
            rtn = conn.update(coll, recs, keyflds)
        return rtn

    def upsert(self, coll, recs, keyflds=None):
        with self.get_connection() as conn:
            rtn = conn.upsert(coll, recs, keyflds)
        return rtn
    
    def delete(self, coll, recs, keyflds=None):
        with self.get_connection() as conn:
            rtn = conn.delete(coll, recs, keyflds)
        return rtn

    def count(self, coll, conds=None):
        with self.get_connection() as conn:
            rtn = conn.count(coll, conds)
        return rtn        

    def drop_collection(self, coll):
        return self.client[self.dbname].drop_collection(coll)

    def __repr__(self):
        return '<Database host={}:{}>'.format(self.dbcfg['host'], self.dbcfg['port'])

    def __enter__(self):
        return self

    def __exit__(self, exc, val, traceback):
        self.close()

    @classmethod
    def id2time(cls, objid):
        objid = str(objid)
        timestamp = int(objid[:8], 16)
        return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(timestamp))

class ZWMongoConnection(object):
    def __init__(self, db):
        self._db = db
        self._cursor = None
        self.open = True
    
    def close(self):
        self._close_cursor()
        self.open = False

    def _close_cursor(self):
        if self._cursor:
            self._cursor.close()
            self._cursor = None

    def __enter__(self):
        return self

    def __exit__(self, exc, val, traceback):
        self.close()

    def __next__(self):
        rec = None
        if self._cursor:
            rec = self._cursor.next()
        if rec:
            return rec
        else:
            self._close_cursor()
            raise StopIteration('Cursor contains no more docs.')

    def find(self, coll, conds=None, projection=None, sort=None, limit=0, fetchall=False, **params):
        sort = sort or [('_id', -1)]
        self._cursor = self._db[coll].find(filter=conds, projection=projection, sort=sort, limit=limit, **params)
        results = DocumentCollection(self)
        if fetchall:
            results.all()
        return results

    def insert(self, coll, recs, ordered=False):
        if recs is None or len(recs) == 0:
            return 0
        result = self._db[coll].insert_many(recs, ordered=ordered)
        return len(result.inserted_ids)
    
    def update(self, coll, recs, keyflds=None):
        if recs is None or len(recs) == 0:
            return 0
        filter_arr = []
        keyflds = keyflds or []
        if '_id' in keyflds:
            reqs = [ UpdateOne({'_id': o['_id']}, {'$set': o}, upsert=False) for i,o in enumerate(recs)]
        else:
            for rec in recs:
                conds = []
                for fld in keyflds:
                    conds.append({fld: rec[fld]})
                filter_arr.append(conds)
            reqs = [ UpdateOne({'$and': filter_arr[i]}, {'$set': o}, upsert=False) for i,o in enumerate(recs)]
        result = self._db[coll].bulk_write(reqs)
        return result.modified_count
    
    def upsert(self, coll, recs, keyflds=None):
        if recs is None or len(recs) == 0:
            return 0
        filter_arr = []
        keyflds = keyflds or []
        if '_id' in keyflds:
            reqs = [ UpdateOne({'_id': o['_id']}, {'$set': o}, upsert=True) for i,o in enumerate(recs)]
        else:
            for rec in recs:
                conds = []
                for fld in keyflds:
                    conds.append({fld: rec[fld]})
                filter_arr.append(conds)
            reqs = [ UpdateOne({'$and': filter_arr[i]}, {'$set': o}, upsert=True) for i,o in enumerate(recs)]
        result = self._db[coll].bulk_write(reqs)
        return result.modified_count, result.upserted_count
    
    def delete(self, coll, recs, keyflds=None):
        if recs is None or len(recs) == 0:
            return 0
        filter_arr = []
        keyflds = keyflds or []
        if '_id' in keyflds:
            reqs = [ DeleteOne({'_id': o['_id']}) for i,o in enumerate(recs)]
        else:
            for rec in recs:
                conds = []
                for fld in keyflds:
                    conds.append({fld: rec[fld]})
                filter_arr.append(conds)
            reqs = [ DeleteOne({'$and': filter_arr[i]}) for i,o in enumerate(recs)]
        result = self._db[coll].bulk_write(reqs)
        return result.deleted_count        

    def count(self, coll, conds=None):
        conds = conds or {}
        return self._db[coll].count_documents(conds)


# -*- coding: utf-8 -*-
import pytest
import os
import sys
TEST_DIR = os.path.abspath(os.path.dirname(__file__))
PARENT_DIR = os.path.join(TEST_DIR, '..')
sys.path.insert(0, PARENT_DIR)

from zwdb import utils

@pytest.mark.parametrize(
    'db_url, result', (
        ('mysql://tester:test@localhost', True),
        ('mysql://tester:test@localhost/testdb', True),
        ('mysql://tester:test@localhost:3306/testdb?useUnicode=true&characterEncoding=UTF-8', True),
    )
)
def test_db_url_parser(db_url, result):
    rtn = utils.db_url_parser(db_url)
    assert (rtn is not None) == result


# -*- coding: utf-8 -*-
import pytest
import os
import sys
from datetime import datetime

from zwdb.zwmongo import ZWMongo

@pytest.fixture(scope='module')
def db():
    db_url = 'mongo://tester:test@localhost/testdb'
    with ZWMongo(db_url, maxPoolSize=50) as mydb:
        recs = [{'txt':'a', 'num':1.0, 'none':None},{'txt':'b', 'num':2.0, 'none':None}]
        mydb.insert('col', recs)
        yield mydb
        colls = ['col', 'col_insert']
        for coll in colls:
            mydb.drop_collection(coll)

class TestMongo:
    def test_list(self, db):
        nms = db.lists()
        assert len(nms)>0

    @pytest.mark.parametrize(
        'recs', (
            [{'id': i, 'txt':'a', 'num':i, 'none':None} for i in range(10)],
        )
    )    
    def test_insert(self, db, recs):
        coll = 'col_insert'
        c = db.insert(coll, recs)
        assert c == len(recs)
    
    def test_find(self, db):
        coll = 'col_insert'
        rs = db.find(coll)
        assert rs.pending == True
        recs = list(rs)
        assert rs.pending == False and recs[0].num == 9 and len(recs) == 10

        rs = db.find(coll)
        assert rs.pending == True
        recs = rs.all()
        assert rs.pending == False and recs[0].num == 9 and len(recs) == 10
  
        rs = db.find(coll)
        assert rs.pending == True
        recs = [r for r in rs]
        assert rs.pending == False and recs[0].num == 9 and len(recs) == 10
        
        rs = db.find(coll, fetchall=True)
        assert rs.pending == False and len(rs) == 10

        rs = db.find(coll, conds={'id':0, 'txt':'a'}, fetchall=True)
        assert len(rs) == 1

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':0, 'txt':'aa'},{'id':1, 'txt':'bb'}],['id']),
        )
    )
    def test_update(self, db, recs, keyflds):
        coll = 'col_insert'
        c = db.update(coll, recs, keyflds=keyflds)
        for rec in recs:
            r = db.find(coll, conds={'id': rec['id']}, fetchall=True)
            assert r[0].txt == rec['txt']
        assert c == len(recs)

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':0, 'txt':'aaa'},{'id':11, 'txt':'a', 'num':11}],['id']),
        )
    )    
    def test_upsert(self, db, recs, keyflds):
        coll = 'col_insert'
        c = db.upsert(coll, recs, keyflds=keyflds)
        total_recs = db.find(coll, fetchall=True)
        for rec in recs:
            r = db.find(coll, conds={'id': rec['id']}, fetchall=True)
            assert r[0].txt == rec['txt']
        assert c[0]+c[1] == len(recs) and 11 == len(total_recs)

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':1},{'id':11}], ['id']),
        )
    )
    def test_delete(self, db, recs, keyflds):
        coll = 'col_insert'
        c = db.delete(coll, recs, keyflds=keyflds)
        total_recs = db.find(coll, fetchall=True)
        assert c == 2 and 9 == len(total_recs)
    
    def test_count(self, db):
        coll = 'col_insert'
        c = db.count(coll)
        assert c == 9


__version__ = '0.0.4'
__title__ = 'zwdb'
__description__ = 'Personal Database Wrapper.'
__url__ = 'https://github.com/njzhaowei'
__author__ = 'Zhao Wei'
__author_email__ = 'yewberry@163.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2020 Zhao Wei'


class Record(object):
    """A row, from a query, from a database."""
    __slots__ = ('_keys', '_values')

    def __init__(self, keys=None, values=None, o=None):
        if isinstance(o, dict):
            self._keys = list(o.keys())
            self._values = list(o.values())
        else:
            self._keys = keys
            self._values = values

        # Ensure that lengths match properly.
        if not isinstance(o, dict):
            assert len(self._keys) == len(self._values)

    def keys(self):
        """Returns the list of column names from the query."""
        return self._keys

    def values(self):
        """Returns the list of values from the query."""
        return self._values

    def __repr__(self):
        return '<Record {}>'.format(self.as_dict())

    def __getitem__(self, key):
        # Support for index-based lookup.
        if isinstance(key, int):
            return self.values()[key]

        # Support for string-based lookup.
        if key in self.keys():
            i = self.keys().index(key)
            if self.keys().count(key) > 1:
                raise KeyError("Record contains multiple '{}' fields.".format(key))
            return self.values()[i]

        raise KeyError("Record contains no '{}' field.".format(key))

    def __getattr__(self, key):
         # Support for attr-based lookup.
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(e)

    def __dir__(self):
        standard = dir(super(Record, self))
        # Merge standard attrs with generated ones (from column names).
        return sorted(standard + [str(k) for k in self.keys()])

    def get(self, key, default=None):
        """Returns the value for a given key, or default."""
        try:
            return self[key]
        except KeyError:
            return default

    def as_dict(self, ordered=False):
        """Returns the row as a dictionary, as ordered."""
        items = zip(self.keys(), self.values())
        return dict(items)

class RecordCollection(object):
    """A set of Records from a query."""
    def __init__(self, keys, rows):
        self._keys = keys
        self._rows = rows
        self._all_rows = []
        self.pending = True

    def all(self, as_dict=False):
        """Returns a list of all rows for the RecordCollection. If they haven't
        been fetched yet, consume the iterator and cache the results."""

        # By calling list it calls the __iter__ method
        rows = list(self)        
        if as_dict:
            return [r.as_dict() for r in rows]
        return rows

    def as_dict(self):
        return self.all(as_dict=True)

    def __iter__(self):
        """Iterate over all rows, consuming the underlying generator only when necessary."""
        i = 0
        while True:
            # Other code may have iterated between yields,
            # so always check the cache.
            if i < len(self):
                yield self[i]
            else:
                # Throws StopIteration when done.
                # Prevent StopIteration bubbling from generator, following https://www.python.org/dev/peps/pep-0479/
                try:
                    yield next(self)
                except StopIteration:
                    return
            i += 1

    def __next__(self):
        try:
            nextrow = next(self._rows)
            nextrec = Record(self._keys, nextrow)
            self._all_rows.append(nextrec)
            return nextrec
        except StopIteration:
            self.pending = False
            raise StopIteration('RecordCollection contains no more rows.')

    def __getitem__(self, key):
        is_int = isinstance(key, int)

        # Convert RecordCollection[1] into slice.
        if is_int:
            key = slice(key, key + 1)

        while len(self) < key.stop or key.stop is None:
            try:
                next(self)
            except StopIteration:
                break

        rows = self._all_rows[key]
        if is_int:
            return rows[0]
        else:
            return RecordCollection(self._keys, iter(rows))

    def __len__(self):
        return len(self._all_rows)

    def __repr__(self):
        return '<RecordCollection size={} pending={}>'.format(len(self), self.pending)

class DocumentCollection(object):
    """A set of Records from a query."""
    def __init__(self, docs):
        self._docs = docs
        self._all_docs = []
        self.pending = True

    def all(self):
        """Returns a list of all rows for the DocumentCollection. If they haven't
        been fetched yet, consume the iterator and cache the results."""

        # By calling list it calls the __iter__ method
        docs = list(self)        
        return docs

    def __iter__(self):
        """Iterate over all rows, consuming the underlying generator only when necessary."""
        i = 0
        while True:
            # Other code may have iterated between yields,
            # so always check the cache.
            if i < len(self):
                yield self[i]
            else:
                # Throws StopIteration when done.
                # Prevent StopIteration bubbling from generator, following https://www.python.org/dev/peps/pep-0479/
                try:
                    yield next(self)
                except StopIteration:
                    return
            i += 1

    def __next__(self):
        try:
            nextdoc = next(self._docs)
            nextrec = Record(o=nextdoc)
            self._all_docs.append(nextrec)
            return nextrec
        except StopIteration:
            self.pending = False
            raise StopIteration('DocumentCollection contains no more docs.')

    def __getitem__(self, key):
        is_int = isinstance(key, int)

        # Convert RecordCollection[1] into slice.
        if is_int:
            key = slice(key, key + 1)

        while len(self) < key.stop or key.stop is None:
            try:
                next(self)
            except StopIteration:
                break

        rows = self._all_docs[key]
        if is_int:
            return rows[0]
        else:
            return DocumentCollection(iter(rows))

    def __len__(self):
        return len(self._all_docs)

    def __repr__(self):
        return '<DocumentCollection size={} pending={}>'.format(len(self), self.pending)

class ZwdbError(Exception):
    def __init__(self, arg):
        self.args = arg


# -*- coding: utf-8 -*-
import pytest
import os
import sys
from datetime import datetime

from zwdb.zwmysql import ZWMysql

@pytest.fixture(scope='module')
def db():
    db_url = 'mysql://tester:test@localhost/testdb'
    with ZWMysql(db_url) as dbobj:
        with dbobj.get_connection() as conn:
            tbl = 'CREATE TABLE `tbl` ( \
                `id` int(11) NOT NULL AUTO_INCREMENT, \
                `txt` varchar(45) DEFAULT NULL, \
                PRIMARY KEY (`id`) \
                ) ENGINE=InnoDB AUTO_INCREMENT=4 DEFAULT CHARSET=utf8mb4;'
            tbl_insert = 'CREATE TABLE `testdb`.`tbl_insert` (\
                `id` INT NOT NULL,\
                `txt` VARCHAR(45) NULL,\
                `num` FLOAT NULL,\
                `none` VARCHAR(45) NULL,\
                `dt` DATETIME NULL,\
                PRIMARY KEY (`id`));'
            tbls = [tbl, tbl_insert]
            for t in tbls:
                conn.execute(t, commit=True)

            recs_insert = [{'id':1, 'txt':'abc'}, {'id':2, 'txt':'def'}, {'id':3, 'txt':'xxx'}]
            ks = recs_insert[0].keys()
            fs = ','.join(ks)
            vs = ','.join(['%({})s'.format(s) for s in ks])
            stmt = 'INSERT INTO tbl ({}) VALUES({})'.format(fs, vs)
            conn.executemany(stmt, fetchall=False, commit=True, paramslist=recs_insert)
        yield dbobj
        # clean
        with dbobj.get_connection() as conn:
            tbls = ['tbl', 'tbl_insert']
            for t in tbls:
                conn.execute('DROP TABLE IF EXISTS %s'%t, commit=True)

class TestMysql:
    def test_list(self, db):
        assert len(db.lists())>0
    
    def test_find(self, db):
        rs = db.find('tbl')
        assert rs.pending == True
        recs = list(rs)
        assert rs.pending == False and recs[0].txt == 'abc' and len(recs) == 3

        rs = db.find('tbl')
        assert rs.pending == True
        recs = rs.all()
        assert rs.pending == False and recs[0].txt == 'abc' and len(recs) == 3
  
        rs = db.find('tbl')
        assert rs.pending == True
        recs = [r for r in rs]
        assert rs.pending == False and recs[0].txt == 'abc' and len(recs) == 3
        
        rs = db.find('tbl', fetchall=True)
        assert rs.pending == False

        rs = db.find('tbl', fetchall=True, id=1, txt='abc')
        assert len(rs) == 1

        rs = db.find('tbl', clause={'order by':'txt desc'}, fetchall=True)
        assert len(rs) == 3

    @pytest.mark.parametrize(
        'recs', (
            [
                {'id':1, 'txt':'a', 'num':1.0, 'none':None, 'dt':datetime.now()},
                {'id':2, 'txt':'b', 'num':2.0, 'none':None, 'dt':datetime.now()}
            ],
        )
    )
    def test_insert(self, db, recs):
        c = db.insert('tbl_insert', recs)
        assert c == len(recs)

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':1, 'txt':'aa'},{'id':2, 'txt':'bb'}],['id']),
        )
    )
    def test_update(self, db, recs, keyflds):
        tbl = 'tbl_insert'
        c = db.update(tbl, recs, keyflds=keyflds)
        for rec in recs:
            r = db.find(tbl, id=rec['id'], fetchall=True)
            assert r[0].txt == rec['txt']
        assert c == len(recs)

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':1, 'txt':'a'},{'id':3, 'txt':'c', 'num':3.0, 'dt':datetime.now()}, {'id':3, 'txt':'haha'}],['id']),
        )
    )    
    def test_upsert(self, db, recs, keyflds):
        tbl = 'tbl_insert'
        c = db.upsert(tbl, recs, keyflds=keyflds)
        total_recs = db.find(tbl, fetchall=True)
        for i,rec in enumerate(recs):
            r = db.find(tbl, id=rec['id'], fetchall=True)
            if i == 1:
                assert r[0].txt == 'haha'
            else:
                assert r[0].txt == rec['txt']
        assert c[0]+c[1] == len(recs) and 3 == len(total_recs)

    @pytest.mark.parametrize(
        'recs, keyflds', (
            ([{'id':1},{'id':3}], ['id']),
        )
    )
    def test_delete(self, db, recs, keyflds):
        tbl = 'tbl_insert'
        c = db.delete('tbl_insert', recs, keyflds=keyflds)
        total_recs = db.find(tbl, fetchall=True)
        assert c == 2 and 1 == len(total_recs)


#!/usr/bin/env python
import os
from setuptools import setup, find_packages
from codecs import open

here = os.path.abspath(os.path.dirname(__file__))

pkg_name = 'zwdb'
packages = [pkg_name]

requires = [s.strip() for s in open('requirements.txt').readlines()]
test_requirements = [s.strip() for s in open('requirements_dev.txt').readlines()][4:]

about = {}
with open(os.path.join(here, pkg_name, '__version__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    packages=packages,
    package_data={'': ['LICENSE', 'NOTICE']},
    package_dir={pkg_name:pkg_name},
    include_package_data=True,
    install_requires=requires,
    tests_require=test_requirements,
    python_requires='>=3.6',
    platforms=["all"],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)